==> scribe-0.1.0-20180129.082643-36.jar: scribe/specs.clj <==
(ns scribe.specs
  "Clojure spec definitions for scribe specs."
  (:require [clojure.spec.alpha :as s]))

;; TODO spec registry
;; TODO registry needs to have fully qualified keys for clj.spec to work

(s/def :scribe/type keyword?)

(defn- type= [t m]
  (= t (:scribe/type m)))

(defn- ->type [type]
  (s/and (s/keys :req [:scribe/type])
         (partial type= type)))

(s/def ::string
  (s/or :wrapped (->type :string)
        :naked #(= :scribe/string %)))

(s/def ::boolean
  (s/or :wrapped (->type :boolean)
        :naked #(= :scribe/boolean %)))

(s/def ::nil
  (s/or :wrapped (->type :nil)
        :naked #(= :scribe/nil %)))

;; TODO ::short (avro "int")
;; TODO ::long (avro "long")
;; TODO ::float (avro "double")
(s/def ::double
  (s/or :wrapped (->type :double)
        :naked #(= :scribe/double %)))

(s/def :scribe/precision integer?)
(s/def :scribe/scale integer?)
(s/def ::decimal
  (s/and (->type :decimal)
         (s/keys :req [:scribe/precision
                       :scribe/scale])))

(s/def ::date (->type :date))
(s/def ::timestamp-micros (->type :timestamp-micros))
(s/def ::timestamp-millis (->type :timestamp-millis))
(s/def ::time-millis (->type :time-millis))
(s/def ::time-micros (->type :time-micros))

(s/def :scribe/enum set?)
(s/def ::enum (s/keys :req [:scribe/enum]))

(s/def :scribe/coll-of
  (s/or :spec-ref keyword?
        :scribe/spec :scribe/spec))
(s/def ::coll-of
  (s/keys :req [:scribe/coll-of]))

(s/def ::keys-coll (s/coll-of keyword?))
(s/def :scribe.keys/req ::keys-coll)
(s/def :scribe.keys/req-un ::keys-coll)
(s/def :scribe.keys/opt ::keys-coll)
(s/def :scribe.keys/opt-un ::keys-coll)
(s/def :scribe/keys
  (s/and (s/keys ::opt #{:scribe.keys/req
                         :scribe.keys/req-un
                         :scribe.keys/opt
                         :scribe.keys/opt-un})
         #(some #{:req :req-un :opt :opt-un} (keys %))))
(s/def ::keys
  (s/and (s/keys :req [:scribe/keys])))

;; TODO make keywords be fully qualified ???
(s/def :scribe/or (s/coll-of keyword?))
(s/def ::or (s/keys :req [:scribe/or]))

(s/def :scribe/spec
  (s/or :scribe/string ::string
        :scribe/boolean ::boolean
        :scribe/nil ::nil
        :scribe/double ::double
        ;;;
        :scribe/decimal ::decimal
        :scribe/date ::date
        :scribe/timestamp-millis ::timestamp-millis
        :scribe/timestamp-micros ::timestamp-micros
        :scribe/time-millis ::time-millis
        :scribe/time-micros ::time-micros
        ;;;
        :scribe/enum ::enum
        :scribe/coll-of ::coll-of
        :scribe/keys ::keys
        :scribe/or ::or))
(s/def ::spec :scribe/spec)

(defn variant
  "Returns the variant of the given scribe spec

  (e.g. :string, :decimal, :time-millis, :coll-of, etc.)"
  [x]
  (when (s/valid? ::spec x)
    (first (s/conform ::spec x))))

==> scribe-0.1.0-20180216.233115-37.jar: scribe/specs.clj <==
(ns scribe.specs
  "Clojure spec definitions for scribe specs."
  (:require [clojure.spec.alpha :as s]))

;; TODO spec registry
;; TODO registry needs to have fully qualified keys for clj.spec to work

(s/def :scribe/type keyword?)

(defn- type= [t m]
  (= t (:scribe/type m)))

(defn- ->type [type]
  (s/and (s/keys :req [:scribe/type])
         (partial type= type)))

(s/def ::string
  (s/or :wrapped (->type :string)
        :naked #(= :scribe/string %)))

(s/def ::boolean
  (s/or :wrapped (->type :boolean)
        :naked #(= :scribe/boolean %)))

(s/def ::nil
  (s/or :wrapped (->type :nil)
        :naked #(= :scribe/nil %)))

;; TODO ::short (avro "int")
;; TODO ::long (avro "long")
;; TODO ::float (avro "double")
(s/def ::double
  (s/or :wrapped (->type :double)
        :naked #(= :scribe/double %)))

(s/def :scribe/precision integer?)
(s/def :scribe/scale integer?)
(s/def ::decimal
  (s/and (->type :decimal)
         ;; TODO make scale and precision optional, only required for avro
         ;; TODO namespace precision and scale further
         (s/keys :req [:scribe/precision
                       :scribe/scale])))

(s/def ::date (->type :date))
(s/def ::timestamp-micros (->type :timestamp-micros))
(s/def ::timestamp-millis (->type :timestamp-millis))
(s/def ::time-millis (->type :time-millis))
(s/def ::time-micros (->type :time-micros))

(s/def :scribe.enum/keyword keyword?)
(s/def :scribe.enum/keywords
  (s/and (s/coll-of :scribe.enum/keyword
                    :kind set?)
         not-empty))
(defn set->wrapped-enum [s]
  {:scribe/type :enum
   :scribe.enum/keywords s})
;; TODO make naked set of keywords valid enum
(s/def ::enum
  (s/and (->type :enum)
         (s/keys :req [:scribe.enum/keywords])))

(s/def :scribe/coll-of
  (s/or :spec-ref keyword?
        :scribe/spec :scribe/spec))
(s/def ::coll-of
  (s/keys :req [:scribe/coll-of]))

(s/def ::keys-coll (s/coll-of keyword?))
(s/def :scribe.keys/req ::keys-coll)
(s/def :scribe.keys/req-un ::keys-coll)
(s/def :scribe.keys/opt ::keys-coll)
(s/def :scribe.keys/opt-un ::keys-coll)
(s/def :scribe/keys
  (s/and (s/keys ::opt #{:scribe.keys/req
                         :scribe.keys/req-un
                         :scribe.keys/opt
                         :scribe.keys/opt-un})
         #(some #{:req :req-un :opt :opt-un} (keys %))))
(s/def ::keys
  (s/and (s/keys :req [:scribe/keys])))

;; TODO make keywords be fully qualified ???
(s/def :scribe/or (s/coll-of keyword?))
(s/def ::or (s/keys :req [:scribe/or]))

(s/def :scribe/spec
  (s/or :scribe/string ::string
        :scribe/boolean ::boolean
        :scribe/nil ::nil
        :scribe/double ::double
        ;;;
        :scribe/decimal ::decimal
        :scribe/date ::date
        :scribe/timestamp-millis ::timestamp-millis
        :scribe/timestamp-micros ::timestamp-micros
        :scribe/time-millis ::time-millis
        :scribe/time-micros ::time-micros
        ;;;
        :scribe/enum ::enum
        :scribe/coll-of ::coll-of
        :scribe/keys ::keys
        :scribe/or ::or))
(s/def ::spec :scribe/spec)

(s/def :scribe/registry
  (s/map-of qualified-keyword?
            :scribe/spec))
(s/def ::registry :scribe/registry)

(defn variant
  "Returns the variant of the given scribe spec

  (e.g. :string, :decimal, :time-millis, :coll-of, etc.)"
  [x]
  (when (s/valid? ::spec x)
    (first (s/conform ::spec x))))
